/* Controller Input Adapter (CIA)
 * Created by Ezechel ("Zeke") Barsan
 * This file uses the example "PS4USB" from the USB_HOST_Library_SAMD folder on github
 * Link to PS4USB: https://github.com/gdsports/USB_Host_Library_SAMD/tree/master/examples/PS4USB
 * Last Update May 26th, 2020
 *  
 * The hdk_generic_data.cpp and hdk_generic_data.h are revisions of coding provided by Tundra labs.
 * 
 * Pinout...
 * Arduino Zero pin ---- Connects to ----- Pin on HDK
 * A0 and A2 both connect to Chip Slave Select 
 * MISO connects to Master Input Slave Output
 * MOSI connects to Master Output Slave Input
 * SLK connects to clock
 * Ground connects to ground
 * 
 * Update 5/26/2020
 * removed the serial print lines that were used for debug purposes
 * commented out the touchpad input/output
 * 
 * 
 */


 
  #include <SPI.h>        //included with arduino IDE
  #include "hdk_generic_data.h" 
  #include <PS4USB.h>     //source - https://github.com/gdsports/USB_Host_Library_SAMD
  
    
/*  This is for debugging purposes, if using Serial
 *#if (USB_VID==0x2341 && defined(ARDUINO_SAMD_ZERO)) || (USB_VID==0x2a03 && defined(ARDUINO_SAM_ZERO))
  #define SerialDebug Serial
  #else
  #define SerialDebug Serial1
  #endif
*/

USBHost UsbH;
PS4USB PS4(&UsbH);

bool printAngle, printTouch;
uint8_t oldL2Value, oldR2Value;

#define CSN     A0

uint8_t x = 0;
uint8_t y = 0;

hdk_generic_data  hdk(4);

volatile uint8_t dmadone;

uint16_t pair_req_cnt = 0;
uint8_t  trigger = 0;

void spi_ss_change(void){
  hdk.ss_deassert_irq();
};


struct buttons {
                uint8_t up : 1;
                uint8_t right : 1;
                uint8_t down : 1;
                uint8_t  left: 1;
                
                uint8_t square : 1;
                uint8_t cross : 1;
                uint8_t circle : 1;

                uint8_t triangle : 1;
                uint8_t l1 : 1;
                uint8_t r1 : 1;
                uint8_t l2 : 1;

                uint8_t r2 : 1;
                uint8_t share : 1;
                uint8_t options : 1;
                uint8_t l3 : 1;

                uint8_t r3 : 1;
                uint8_t ps : 1;
} __attribute__((packed));

struct HatXY {
    uint8_t x;
    uint8_t y;
};

/*
 * This struct is for the touchpad coordinates, not working at the moment.
struct tpXY {
    uint16_t x : 12;
    uint16_t y : 12;
} __attribute__((packed));
*/

typedef struct __attribute__((packed, aligned(1)))
{
    buttons  btn;
    HatXY       hat[2];
   // tpXY  touchpad[2]; not using touchpad data in this version 
} ps4_dataframe_t;



ps4_dataframe_t cdata;


void hdk_event_cb( uint8_t event_id, uint8_t payload_len, char* payload ) {
    switch (event_id) {
      case MI_MASTER_EVENT_WIRLESS_PAIRING_STARTED:
       break;
      case MI_MASTER_EVENT_WIRLESS_PAIRING_STOPPED:
        break;
      case GDATA_EVENT_PROTOCOL_REVISON_CHANGED:
       break;
      case MI_MASTER_EVENT_MFG_CFG_RESP:
       payload[payload_len] = 0;
        break;
      case MI_MASTER_EVENT_HAPTIC_PULSE:
         payload[payload_len] = 0;
         // Insert PS4 Rumble here
        break;
      case MI_MASTER_EVENT_HAPTIC_SONG:
       break;
      case MI_MASTER_EVENT_SHUTDOWN_WARNING:
        break;
      case MI_MASTER_EVENT_ON_CHARGER:
        break;
      case MI_MASTER_EVENT_OFF_CHARGER:
         break;
      case MI_MASTER_EVENT_POWER_GOOD:
        break;
      case MI_MASTER_EVENT_POWER_NOT_GOOD:
        break;
      case MI_MASTER_EVENT_USB_HOST:
        break;
      case MI_MASTER_EVENT_USB_CHARGER:
        break;
      case MI_MASTER_EVENT_USB_UNKNOWN:
       break;
      case MI_MASTER_EVENT_FIRMWARE_REV_REQ:
        hdk.send_event(MI_SLAVE_EVENT_FIRMWARE_REV_RESP);
        hdk.send_event(1);
        hdk.send_event(2);
        hdk.send_event(3);
        hdk.send_event(4);
        break;
      default:
        break;
    }
    
    if (event_id == MI_MASTER_EVENT_HAPTIC_PULSE) 
    {
      mi_haptic_event_t* haptic = (mi_haptic_event_t*) payload;
     }
}

void setup() {
  
  pinMode(A0, INPUT_PULLUP);
  
 /* Serial.begin(115200);
  while(!Serial);
  Serial.println("Tundra Labs HDK Generic Data Test");
  delay(30);
 */  
 
  hdk.init();
  hdk.register_event_callback(hdk_event_cb);
  attachInterrupt(A0, spi_ss_change, RISING); 
  if (UsbH.Init()) {
  //  SerialDebug.print(F("\r\nUSB host did not start"));
    while (1); // Halt
  }
 // SerialDebug.print(F("\r\nPS4 USB Library Started"));
  delay(30);
}

//uint8_t backchannel_buf[17];
//uint8_t backchannel_len = 0;

void loop() { 
  delay(20);
  UsbH.Task();
        
  cdata.btn.up = PS4.getButtonClick(UP);
  cdata.btn.right = PS4.getButtonClick(RIGHT);
  cdata.btn.down = PS4.getButtonClick(DOWN);
  cdata.btn.left = PS4.getButtonClick(LEFT);
  
  cdata.btn.square = PS4.getButtonClick(SQUARE);
  cdata.btn.cross = PS4.getButtonClick(CROSS);
  cdata.btn.circle = PS4.getButtonClick(CIRCLE);
  cdata.btn.triangle = PS4.getButtonClick(TRIANGLE);
  
  cdata.btn.l1 = PS4.getButtonClick(L1);
  cdata.btn.r1 = PS4.getButtonClick(R1);
  cdata.btn.l2 = PS4.getAnalogButton(L2);
  cdata.btn.r2 = PS4.getAnalogButton(R2);

  cdata.btn.share = PS4.getButtonClick(SHARE);
  cdata.btn.options = PS4.getButtonClick(OPTIONS);

  cdata.btn.l3 = PS4.getButtonClick(L3);
  cdata.btn.r3 = PS4.getButtonClick(R3);
  cdata.btn.ps = PS4.getButtonClick(PS);

  cdata.hat[0].x = PS4.getAnalogHat(LeftHatX);
  cdata.hat[0].y = PS4.getAnalogHat(LeftHatY);
  cdata.hat[1].x = PS4.getAnalogHat(RightHatX);
  cdata.hat[1].y = PS4.getAnalogHat(RightHatY);
  
  /*
   * The touchpad has 3 angles, according to ps4parser.h file in the USBPS4 library on github
   * will need further investigation.
   * Also the touchpad can be clicked on, so could be a button - just a thought
  cdata.touchpad[0].x = PS4.getX(0);
  cdata.touchpad[0].y = PS4.getY(0); 
  cdata.touchpad[1].x = PS4.getX(1);
  cdata.touchpad[1].y = PS4.getY(1); 
  */
  
  hdk.write_data(0,(uint8_t *)   &cdata, sizeof(cdata));
  
}

void SERCOM4_Handler(void) {
  hdk.sercom_irq();
  //attachInterrupt(A0, spi_ss_change, CHANGE);
}
