#include <Arduino.h>
#include "hdk_generic_data.h"

// Create a lookup table for the payload length for events, if the event is not listed,
// The LUT value is automatically 0
/*
  static const uint8_t mi_master_event_payload_length[MI_MASTER_NUM_EVENTS] =
  {
  [MI_MASTER_EVENT_MFG_CFG_RESP] = 32 ,
  [MI_MASTER_EVENT_HAPTIC_PULSE] = 14
  };

  static const uint8_t mi_slave_event_payload_length[MI_SLAVE_NUM_EVENTS] =
  {
  [MI_SLAVE_EVENT_FIRMWARE_REV_RESP] = 8,
  };
*/
hdk_generic_data::hdk_generic_data(int sercom_num) {
  sercom = sercom_num;
  active_buff = 0;
  buffer_mutex = true;
  protocol_revision = 0;
  event_cb = 0;
}

void hdk_generic_data::init(void) {

  id_ptr = 0;
 // Serial.println("init dma");
  //delay(30);
  initDma();
  //Serial.println("done");
  //delay(30);

  disableDmaChannel(0);
  configureTxDma(0, &tx_buf[0], sizeof(tx_buf[0]));

  disableDmaChannel(1);
  configureTxDma(1, &tx_buf[1], sizeof(tx_buf[1]));

  disableDmaChannel(2);
  configureRxDma(2, &rx_buf[0], sizeof(rx_buf[0]));

  disableDmaChannel(3);
  configureRxDma(3, &rx_buf[1], sizeof(rx_buf[1]));

  enableDmaChannel(0);
  enableDmaChannel(2);

  tx_buf[0].header.spi_protocol_rev = 0x01;
  tx_buf[1].header.spi_protocol_rev = 0x01;

 // Serial.println("init sercom");
 // delay(30);
  initSercom();
}

uint8_t hdk_generic_data::get_protocol_rev(void) {
  return protocol_revision;
}

void DMAC_Handler() {
  uint8_t active_channel = DMAC->INTPEND.bit.ID;
  __disable_irq();
  DMAC->CHID.reg = DMAC_CHID_ID(active_channel);
  //dmadone = DMAC->CHINTFLAG.reg;
  DMAC->CHINTFLAG.reg = DMAC_CHINTENCLR_TCMPL; // clear
  DMAC->CHINTFLAG.reg = DMAC_CHINTENCLR_TERR;
  DMAC->CHINTFLAG.reg = DMAC_CHINTENCLR_SUSP;
  __enable_irq();
}

void hdk_generic_data::sercom_irq(void) {
  if (SERCOM4->SPI.INTFLAG.bit.SSL)
  {

    ss_assert_irq();
    SERCOM4->SPI.INTFLAG.bit.SSL = 1; //clear slave select interrupt
  }
  /*
    else if (int_flags & (1 << 1))
    {
    SERCOM4->SPI.DATA.reg = 0xCC;
    SERCOM4->SPI.INTFLAG.bit.TXC = 1;
    }
    else if (int_flags & (1 << 2))
    {
    data = SERCOM4->SPI.DATA.reg;
    SERCOM4->SPI.DATA.reg = 0xBB;
    SERCOM4->SPI.INTFLAG.bit.RXC = 1;
    }
    else*/
  if (SERCOM4->SPI.INTFLAG.bit.DRE)
  {
    SERCOM4->SPI.INTFLAG.bit.DRE = 1;
  }
};

void hdk_generic_data::ss_assert_irq(void) {
  uint8_t buf_to_send = active_buff;

  active_buff ^= 1;
  tx_buf[active_buff].header.event_data_length = 0;
  tx_buf[active_buff].header.backchannel_length = 0;
  tx_buf[active_buff].header.input_data_length = 0;

  uint8_t buf_ptr = tx_buf[buf_to_send].header.input_data_length;

  if (backchannel_length > 0) {
    memcpy(&tx_buf[buf_to_send].buffer[buf_ptr], backchannel_buf, backchannel_length);
    tx_buf[buf_to_send].header.backchannel_length = backchannel_length;
    buf_ptr += backchannel_length;
    backchannel_length = 0;
  }

  uint8_t event_buff_len = 0;
  while (event_queue_pull_i != event_queue_push_i && (tx_buf[buf_to_send].header.input_data_length + event_buff_len) < MI_MAXIMUM_FRAME_BUFFER_LEN) {
    tx_buf[buf_to_send].buffer[buf_ptr++] = event_queue[event_queue_pull_i];
    event_queue_pull_i = (event_queue_pull_i + 1) & 0x0F;

    tx_buf[buf_to_send].header.event_data_length++;
  }

}

void hdk_generic_data::ss_deassert_irq(void) {

  disableDmaChannel(active_buff ^ 1);
  disableDmaChannel(2 + active_buff ^ 1);

  SERCOM4->SPI.CTRLA.bit.ENABLE = 0;

  tx_buf[0].header.frame_id++;
  tx_buf[1].header.frame_id++;

  enableDmaChannel(2 + (active_buff));
  enableDmaChannel(active_buff);

  SERCOM4->SPI.CTRLA.bit.ENABLE = 1;

  if  ( tx_buf[active_buff].header.spi_protocol_rev == 0x01 )
  {
    uint8_t revision = rx_buf[active_buff].header.report_mode_request;

    if (revision != protocol_revision) {
      protocol_revision = revision;
      tx_buf[0].header.report_mode = protocol_revision;
      tx_buf[1].header.report_mode = protocol_revision;
      event_cb(GDATA_EVENT_PROTOCOL_REVISON_CHANGED, 0, NULL);
    }
    // Check for events
    uint8_t data_index = 0;

    if ( rx_buf[active_buff].header.status.enter_bootloader == 1 )
    {
      if ( tx_buf[active_buff].header.status.bootloader != 1 ) {
       // Serial.println("Entering Bootloader Mode\n");
        tx_buf[0].header.status.bootloader = 1;
        tx_buf[1].header.status.bootloader = 1;
      }
    }
    else {
      if ( tx_buf[active_buff].header.status.bootloader == 1 )
      {
       // Serial.println("Exiting Bootloader Mode\n");
        tx_buf[0].header.status.bootloader = 0;
        tx_buf[1].header.status.bootloader = 0;
      }
    }

    while (data_index < rx_buf[active_buff].header.backchannel_length) {
    //  Serial.print((char) rx_buf[active_buff].buffer[data_index]);
      data_index++;
    }
    if (rx_buf[active_buff].header.backchannel_length != 0)
    {
     // Serial.println(" ");
    }

    uint8_t evt_index = 0;
    while (evt_index < rx_buf[active_buff].header.event_data_length) {
      uint8_t evt_payload_length = 0;
      if (event_cb) {
        uint8_t evt_opcode = rx_buf[active_buff].buffer[data_index];
        //uint8_t evt_payload_length = mi_master_event_payload_length[evt_opcode];

        switch (evt_opcode)
        {
          case MI_MASTER_EVENT_HAPTIC_PULSE:
            evt_payload_length = 7;
            break;
          case MI_MASTER_EVENT_HAPTIC_SONG:
            evt_payload_length = 1;
            break;
        }
        if (evt_payload_length > 0) {
          event_cb(evt_opcode, evt_payload_length, (char*) &rx_buf[active_buff].buffer[data_index + 1]);
        }
        else {
          event_cb(evt_opcode, 0, NULL);
        }
      }
      else {
        break;
      }
      evt_index = evt_payload_length + 1;
      data_index++;
    }
  }
}

void hdk_generic_data::update(void) {
  uint8_t ss = ((PORT->Group[PORTB].IN.reg) & (1 << 9));
  if (ss !=  last_ss) {
    PORT->Group[PORTA].OUTSET.reg = 1 << 15;
    last_ss = ss;
    PORT->Group[PORTA].OUTCLR.reg = 1 << 15;
  }
}

void hdk_generic_data::write_data(uint8_t offset, uint8_t d) {
  tx_buf[0].buffer[offset] = d;
  tx_buf[0].header.input_data_length = (offset + 1) > tx_buf[0].header.input_data_length ? (offset + 1) : tx_buf[0].header.input_data_length;
  tx_buf[1].buffer[offset] = d;
  tx_buf[1].header.input_data_length = (offset + 1) > tx_buf[1].header.input_data_length ? (offset + 1) : tx_buf[1].header.input_data_length;
}

void hdk_generic_data::write_data(uint8_t offset, uint8_t *src, uint8_t len)
{
  memcpy(&tx_buf[0].buffer[offset], src, len);
  memcpy(&tx_buf[1].buffer[offset], src, len);
  tx_buf[0].header.input_data_length = offset + len;
  tx_buf[1].header.input_data_length = offset + len;
}

void hdk_generic_data::send_event(uint8_t event_id) {
  event_queue[event_queue_push_i] = event_id;
  event_queue_push_i = (event_queue_push_i + 1) & 0x0F;
}

bool hdk_generic_data::backchannel_available( void )
{
  return (backchannel_length == 0);
}

void hdk_generic_data::write_backchannel(uint8_t *buf, uint8_t len)
{
  memcpy(backchannel_buf, buf, len);
  backchannel_length = len;
}

void hdk_generic_data::register_event_callback(hdk_event_callback_t cb) {
  event_cb = cb;
}

void hdk_generic_data::initDma(void) {
  PM->AHBMASK.reg |= PM_AHBMASK_DMAC ;
  PM->APBBMASK.reg |= PM_APBBMASK_DMAC ;
  NVIC_EnableIRQ( DMAC_IRQn ) ;

  DMAC->BASEADDR.reg = (uint32_t)descriptor_section;
  DMAC->WRBADDR.reg = (uint32_t)wrb;
  DMAC->CTRL.reg = DMAC_CTRL_DMAENABLE | DMAC_CTRL_LVLEN(0xf);
};

void hdk_generic_data::enableDmaChannel(uint8_t chnl) {
  DMAC->CHID.reg = DMAC_CHID_ID(chnl);
  DMAC->CHCTRLA.reg |= DMAC_CHCTRLA_ENABLE;
};

void hdk_generic_data::disableDmaChannel(uint8_t chnl) {
  DMAC->CHID.reg = DMAC_CHID_ID(chnl);
  DMAC->CHCTRLA.reg &= ~DMAC_CHCTRLA_ENABLE;
};

void hdk_generic_data::configureRxDma(uint32_t chnl, void *rxdata,  size_t hwords) {
  DMAC->CHID.reg = DMAC_CHID_ID(chnl);
  DMAC->CHCTRLA.reg = DMAC_CHCTRLA_SWRST;
  DMAC->SWTRIGCTRL.reg &= (uint32_t)(~(1 << chnl));
  DMAC->CHCTRLB.reg = DMAC_CHCTRLB_LVL(0) | DMAC_CHCTRLB_TRIGSRC(SERCOM4_DMAC_ID_RX) | DMAC_CHCTRLB_TRIGACT_BEAT;
  DMAC->CHINTENSET.reg = DMAC_CHINTENSET_MASK ;

  descriptor.descaddr = 0;
  descriptor.srcaddr = (uint32_t) &SERCOM4->SPI.DATA.reg;
  descriptor.btcnt =  hwords;
  descriptor.dstaddr = (uint32_t)rxdata + hwords;   // end address
  descriptor.btctrl =  DMAC_BTCTRL_BEATSIZE_BYTE | DMAC_BTCTRL_DSTINC | DMAC_BTCTRL_VALID;

  memcpy(&descriptor_section[chnl], &descriptor, sizeof(dmacdescriptor));
}

void hdk_generic_data::configureTxDma(uint32_t chnl, void *txdata,  size_t hwords) {
  DMAC->CHID.reg = DMAC_CHID_ID(chnl);
  DMAC->CHCTRLA.reg = DMAC_CHCTRLA_SWRST;
  DMAC->SWTRIGCTRL.reg &= (uint32_t)(~(1 << chnl));
  DMAC->CHCTRLB.reg = DMAC_CHCTRLB_LVL(1) | DMAC_CHCTRLB_TRIGSRC(SERCOM4_DMAC_ID_TX) | DMAC_CHCTRLB_TRIGACT_BEAT;
  DMAC->CHINTENSET.reg = DMAC_CHINTENSET_MASK;

  descriptor.descaddr = (uint32_t) &descriptor_section[chnl]; // set to 0 for oneshot
  descriptor.srcaddr = (uint32_t)txdata + hwords;
  descriptor.btcnt =  hwords;
  descriptor.dstaddr = (uint32_t) &SERCOM4->SPI.DATA.reg;   // end address
  descriptor.btctrl =  DMAC_BTCTRL_BEATSIZE_BYTE | DMAC_BTCTRL_STEPSEL_SRC | DMAC_BTCTRL_SRCINC | DMAC_BTCTRL_VALID;

  memcpy(&descriptor_section[chnl], &descriptor, sizeof(dmacdescriptor));
}

void hdk_generic_data::initSercom(void) {
  PORT->Group[PORTB].PINCFG[11].bit.PMUXEN = 0x1; //Enable Peripheral Multiplexing for SERCOM4 SPI PA18 Arduino PIN10
  PORT->Group[PORTB].PMUX[5].bit.PMUXO = 0x3; //SERCOM 1 is selected for peripherial use of this pad

  PORT->Group[PORTB].PINCFG[10].bit.PMUXEN = 0x1; //Enable Peripheral Multiplexing for SERCOM4 SPI PA18 Arduino PIN10
  PORT->Group[PORTB].PMUX[5].bit.PMUXE = 0x3; //SERCOM 1 is selected for peripherial use of this pad
  PORT->Group[PORTA].PINCFG[12].bit.PMUXEN = 0x1; //Enable Peripheral Multiplexing for SERCOM4 SPI PA18 Arduino PIN10
  PORT->Group[PORTA].PMUX[6].bit.PMUXE = 0x3; //SERCOM 1 is selected for peripherial use of this pad
  PORT->Group[PORTB].PINCFG[9].bit.PMUXEN = 0x1; //Enable Peripheral Multiplexing for SERCOM4 SPI PA18 Arduino PIN10
  PORT->Group[PORTB].PMUX[4].bit.PMUXO = 0x3; //SERCOM 1 is selected for peripherial use of this pad
  //Disable SPI 1
  SERCOM4->SPI.CTRLA.bit.ENABLE = 0;
  while (SERCOM4->SPI.SYNCBUSY.bit.ENABLE);
  //Reset SPI 1
  SERCOM4->SPI.CTRLA.bit.SWRST = 1;
  while (SERCOM4->SPI.CTRLA.bit.SWRST || SERCOM4->SPI.SYNCBUSY.bit.SWRST);
  //Setting up NVIC
  NVIC_EnableIRQ(SERCOM4_IRQn);
  NVIC_SetPriority(SERCOM4_IRQn, 2);
  //Setting Generic Clock Controller!!!!
  GCLK->CLKCTRL.reg = GCLK_CLKCTRL_ID(GCM_SERCOM4_CORE) | //Generic Clock 0
                      GCLK_CLKCTRL_GEN_GCLK0 | // Generic Clock Generator 0 is the source
                      GCLK_CLKCTRL_CLKEN; // Enable Generic Clock Generator

  while (GCLK->STATUS.reg & GCLK_STATUS_SYNCBUSY); //Wait for synchronisation

  //Set up SPI Control A Register
  SERCOM4->SPI.CTRLA.bit.DORD = 0; //MSB first
  SERCOM4->SPI.CTRLA.bit.CPOL = 0; //SCK is low when idle, leading edge is rising edge
  SERCOM4->SPI.CTRLA.bit.CPHA = 0; //data sampled on leading sck edge and changed on a trailing sck edge
  SERCOM4->SPI.CTRLA.bit.FORM = 0x0; //Frame format = SPI
  SERCOM4->SPI.CTRLA.bit.DIPO = 0x2; //DATA PAD0 MOSI is used as input (slave mode)
  SERCOM4->SPI.CTRLA.bit.DOPO = 0x3; //DATA PAD3 MISO is used as output
  SERCOM4->SPI.CTRLA.bit.MODE = 0x2; //SPI in Slave mode
  SERCOM4->SPI.CTRLA.bit.IBON = 0x1; //Buffer Overflow notification
  SERCOM4->SPI.CTRLA.bit.RUNSTDBY = 1; //wake on receiver complete

  //Set up SPI control B register
  //SERCOM4->SPI.CTRLB.bit.RXEN = 0x1; //Enable Receiver
  SERCOM4->SPI.CTRLB.bit.SSDE = 0x1; //Slave Selecte Detection Enabled
  SERCOM4->SPI.CTRLB.bit.CHSIZE = 0; //character size 8 Bit
  SERCOM4->SPI.CTRLB.bit.PLOADEN = 0x1; //Enable Preload Data Register
  //while (SERCOM4->SPI.SYNCBUSY.bit.CTRLB);
  //Set up SPI interrupts
  SERCOM4->SPI.INTENSET.bit.SSL = 0x1; //Enable Slave Select low interrupt
  SERCOM4->SPI.INTENSET.bit.RXC = 0x0; //Receive complete interrupt
  SERCOM4->SPI.INTENSET.bit.TXC = 0x0; //Receive complete interrupt
  SERCOM4->SPI.INTENSET.bit.ERROR = 0x0; //Receive complete interrupt
  SERCOM4->SPI.INTENSET.bit.DRE = 0x1; //Data Register Empty interrupt
  //init SPI CLK
  //SERCOM4->SPI.BAUD.reg = SERCOM_FREQ_REF / (2*4000000u)-1;
  //Enable SPI
  SERCOM4->SPI.CTRLA.bit.ENABLE = 1;
  //while (SERCOM4->SPI.SYNCBUSY.bit.ENABLE);
  SERCOM4->SPI.CTRLB.bit.RXEN = 0x1; //Enable Receiver, this is done here due to errate issue
  while (SERCOM4->SPI.SYNCBUSY.bit.CTRLB); //wait until receiver is enabled

}
