
#define SC SERCOM4

#define MI_MAXIMUM_FRAME_BUFFER_LEN            (64)

typedef struct {
    uint16_t btctrl;
    uint16_t btcnt;
    uint32_t srcaddr;
    uint32_t dstaddr;
    uint32_t descaddr;
} dmacdescriptor ;

#pragma pack(push)

typedef struct __attribute__((packed, aligned(1)))
{
    uint8_t                 frame_error: 1;
    uint8_t                 unsupported_spi_protocol_rev: 1;
    uint8_t                 upstream_backchannel_full: 1;
    uint8_t                 is_wireless: 1;
    uint8_t                 enter_bootloader: 1;
    uint8_t                 reserved: 3;
} mi_master_status_t;

typedef struct __attribute__((packed, aligned(1)))
{
    uint8_t                 spi_protocol_rev;
    uint8_t                 frame_id;
    uint8_t                 report_mode_request;
    mi_master_status_t      status;
    uint8_t                 backchannel_length; 
    uint8_t                 event_data_length;
    uint8_t                 reserved[2]; 
} mi_master_header_t;

typedef struct __attribute__((packed, aligned(1)))
{
    mi_master_header_t      header;
    uint8_t                 buffer [ MI_MAXIMUM_FRAME_BUFFER_LEN ];
} mi_master_frame_t;

typedef struct __attribute__((packed, aligned(1)))
{
    uint8_t                 bootloader: 1;  
    uint8_t                 reserved: 7;
} mi_slave_status_t;

typedef struct __attribute__((packed, aligned(1)))
{
    uint8_t                 spi_protocol_rev;
    uint8_t                 frame_id;
    uint8_t                 report_mode;
    mi_slave_status_t       status;
    uint8_t                 input_data_length;
    uint8_t                 backchannel_length; 
    uint8_t                 event_data_length;
    uint8_t                 reserved;
} mi_slave_header_t;

typedef struct __attribute__((packed, aligned(1)))
{
    mi_slave_header_t       header;
    uint8_t                 buffer [ MI_MAXIMUM_FRAME_BUFFER_LEN ];
} mi_slave_frame_t;

typedef struct __attribute__((packed, aligned(1)))
{
  uint8_t       system_button: 1;
  uint8_t       menu_button: 1;
  uint8_t       trackpad_button: 1;
  uint8_t       grip_button: 1;
  uint8_t       trigger_button: 1;
  uint8_t       trackpad_finger_present:1;
  uint8_t       reserved: 2;
} mi_legacy_mode_buttons_t;

typedef struct __attribute__((packed, aligned(1)))
{
  uint8_t                   trigger;
  int16_t                   trackpad_x;
  int16_t                   trackpad_y; 
  mi_legacy_mode_buttons_t  buttons; 
} mi_legacy_mode_0x01_input_data_t;

typedef struct __attribute__((packed, aligned(1)))
{
  uint16_t     duration_us;
  uint16_t     count;
  uint16_t     interval_us; 
  uint8_t      priority;
} mi_haptic_event_t;

typedef enum {
  MI_PROTOCOL_REVISON_NOP                       = 0,
  MI_PROTOCOL_REVISON_V1_CONTROLLER             = 1,  // Data structure like the Vive Wand
  MI_PROTOCOL_REVISON_V2_CONTROLLER             = 2,  // Data structure like knuckles (Currently not supported)
  MI_PROTOCOL_REVISON_GENERIC                   = 3,  // Generic data structure
} mi_protocol_revision_t;

typedef enum {
  MI_MASTER_EVENT_NOP                           = 0,
  MI_MASTER_TEST_EVENT_NO_PAYLOAD               = 1,
  MI_MASTER_TEST_EVENT_8B_PAYLOAD               = 2,
  MI_MASTER_EVENT_WIRLESS_PAIRING_STARTED       = 3,
  MI_MASTER_EVENT_WIRLESS_PAIRING_STOPPED       = 4,
  MI_MASTER_EVENT_WIRLESS_PAIRING_CONNECTED     = 5,
  MI_MASTER_EVENT_WIRLESS_PAIRING_DISCONNECTED  = 6,
  MI_MASTER_EVENT_MFG_CFG_RESP                  = 7,
  MI_MASTER_EVENT_FIRMWARE_REV_REQ              = 8,
  MI_MASTER_EVENT_HAPTIC_PULSE                  = 9,
  MI_MASTER_EVENT_HAPTIC_PULSE_REPEAT           = 10,
  MI_MASTER_EVENT_HAPTIC_SONG                   = 11,
  MI_MASTER_EVENT_SHUTDOWN_WARNING              = 12,
  MI_MASTER_EVENT_ON_CHARGER                    = 13,
  MI_MASTER_EVENT_OFF_CHARGER                   = 14,
  MI_MASTER_EVENT_POWER_GOOD                    = 15,
  MI_MASTER_EVENT_POWER_NOT_GOOD                = 16,
  MI_MASTER_EVENT_USB_HOST                      = 17,
  MI_MASTER_EVENT_USB_CHARGER                   = 18,
  MI_MASTER_EVENT_USB_UNKNOWN                   = 19,
  MI_MASTER_NUM_EVENTS,

  GDATA_EVENT_PROTOCOL_REVISON_CHANGED,
} mi_master_event_t;

typedef enum {
  MI_SLAVE_EVENT_NOP                            = 0,
  MI_SLAVE_TEST_EVENT_NO_PAYLOAD                = 1,
  MI_SLAVE_TEST_EVENT_8B_PAYLOAD                = 2,
  MI_SLAVE_EVENT_BEGIN_WIRELESS_PAIRING         = 3,
  MI_SLAVE_EVENT_MFG_CFG_REQ                    = 4,
  MI_SLAVE_EVENT_FIRMWARE_REV_RESP              = 5,
  MI_SLAVE_NUM_EVENTS,
} mi_slave_event_t;

#pragma pack(pop)

typedef void ( *hdk_event_callback_t )( uint8_t event_code,uint8_t payload_len, char* payload );

class hdk_generic_data {
  public:
    hdk_generic_data(int sercom_num);
    void init(void);

    void update(void);

    uint8_t get_protocol_rev(void);

    bool backchannel_available();
    void write_backchannel(uint8_t *buf, uint8_t len);
    
    void ss_assert_irq(void);
    void ss_deassert_irq(void);
    void sercom_irq(void);
    void dma_irq(void);
    void write_data(uint8_t offset, uint8_t d);
    void write_data(uint8_t offset, uint8_t *src, uint8_t len);
    void send_event(uint8_t event_id);
    void register_event_callback(hdk_event_callback_t cb);
  private:  

    uint8_t id_ptr;
    bool dma_activated;
    uint8_t sercom;
    uint8_t active_buff;
    uint8_t last_ss;
    bool buffer_mutex;

    uint8_t  backchannel_length;
    uint8_t  backchannel_buf[32];

    mi_master_frame_t    rx_buf[2];
    mi_slave_frame_t    tx_buf[2];

    uint8_t event_queue[16];
    uint8_t event_queue_push_i;
    uint8_t event_queue_pull_i;

    uint8_t protocol_revision;

    hdk_event_callback_t event_cb;
    
    volatile dmacdescriptor wrb[12] __attribute__ ((aligned (16)));
    dmacdescriptor descriptor_section[12] __attribute__ ((aligned (16)));
    dmacdescriptor descriptor __attribute__ ((aligned (16)));

    void initSercom(void);
    void initDma(void);
    void enableDmaChannel(uint8_t chnl);
    void disableDmaChannel(uint8_t chnl);
    void configureRxDma(uint32_t chnl, void *rxdata,  size_t hwords);
    void configureTxDma(uint32_t chnl, void *txdata,  size_t hwords);
};
